import pytest
from flash.image import ImageClassificationData

from gradsflow.automodel import AutoModel

datamodule = ImageClassificationData.from_folders(
    train_folder="data/hymenoptera_data/train/",
    val_folder="data/hymenoptera_data/val/",
)


def test_auto_model():
    assert AutoModel(datamodule)


def test_build_model():
    model = AutoModel(datamodule)
    with pytest.raises(NotImplementedError):
        model.build_model(**{"lr": 1})


def test_build_model():
    model = AutoModel(datamodule)
    with pytest.raises(NotImplementedError):
        model._objective(None)



from abc import abstractmethod
from typing import Dict, List, Optional, Union

import optuna
import torch
from flash.core.data.data_module import DataModule

from gradsflow.automodel.automodel import AutoModel


# noinspection PyTypeChecker
class AutoClassifier(AutoModel):
    DEFAULT_BACKBONES = []

    def __init__(
        self,
        datamodule: DataModule,
        max_epochs: int = 10,
        n_trials: int = 100,
        optimization_metric: Optional[str] = None,
        suggested_backbones: Union[List, str, None] = None,
        suggested_conf: Optional[dict] = None,
        timeout: int = 600,
        prune: bool = True,
        optuna_confs: Optional[Dict] = None,
    ):
        super().__init__(
            datamodule,
            max_epochs=max_epochs,
            optimization_metric=optimization_metric,
            n_trials=n_trials,
            suggested_conf=suggested_conf,
            timeout=timeout,
            prune=prune,
            optuna_confs=optuna_confs,
        )

        if not suggested_backbones:
            self.suggested_backbones = self.DEFAULT_BACKBONES
        elif isinstance(suggested_backbones, str):
            self.suggested_backbones = [suggested_backbones]
        elif isinstance(suggested_backbones, (list, tuple)):
            self.suggested_backbones = suggested_backbones
        else:
            raise UserWarning("Invalid suggested_backbone type!")

        self.datamodule = datamodule
        self.num_classes = datamodule.num_classes

    def forward(self, x):
        if not self.model:
            raise UserWarning("model not initialized yet!")
        return self.model(x)

    # noinspection PyTypeChecker
    def get_trial_model(self, trial: optuna.Trial) -> Dict[str, str]:

        trial_backbone = trial.suggest_categorical("backbone", self.suggested_backbones)
        trial_lr = trial.suggest_float("lr", *self.suggested_lr, log=True)
        trial_optimizer = trial.suggest_categorical(
            "optimizer", self.suggested_optimizers
        )
        hparams = {
            "backbone": trial_backbone,
            "lr": trial_lr,
            "optimizer": trial_optimizer,
        }
        return hparams

    @abstractmethod
    def build_model(self, **kwargs) -> torch.nn.Module:
        raise NotImplementedError



"""An AutoML Library made with Optuna and PyTorch Lightning"""

from gradsflow.autoclassifier.image import AutoImageClassifier
from gradsflow.autoclassifier.text import AutoTextClassifier
from gradsflow.automodel import AutoModel

__version__ = "0.0.1b1"



from abc import abstractmethod
from typing import Dict, Optional, Union

import optuna
import pytorch_lightning as pl
import torch
from flash import DataModule
from optuna.integration import PyTorchLightningPruningCallback

from gradsflow.logging import logger
from gradsflow.utility.common import module_to_cls_index


class AutoModel:
    """
    Creates Optuna instance and suggested objects from hparams
    """

    OPTIMIZER_INDEX = module_to_cls_index(torch.optim, True)
    DEFAULT_OPTIMIZERS = ["adam", "sgd"]
    DEFAULT_LR = (1e-5, 1e-1)

    def __init__(
        self,
        datamodule: DataModule,
        max_epochs: int = 10,
        optimization_metric: Optional[str] = None,
        n_trials: int = 100,
        suggested_conf: Optional[dict] = None,
        timeout: int = 600,
        prune: bool = True,
        optuna_confs: Optional[Dict] = None,
    ):

        self.pruner: optuna.pruners.BasePruner = (
            optuna.pruners.MedianPruner() if prune else optuna.pruners.NopPruner()
        )
        self.datamodule = datamodule
        self.n_trials = n_trials
        self.model: Union[torch.nn.Module, pl.LightningModule, None] = None
        self.max_epochs = max_epochs
        self.timeout = timeout
        if not optimization_metric:
            optimization_metric = "val_accuracy"
        self.optimization_metric = optimization_metric
        if not optuna_confs:
            optuna_confs = {}
        self.optuna_confs = optuna_confs
        self.study = optuna.create_study(
            optuna_confs.get("storage"),
            pruner=self.pruner,
            study_name=optuna_confs.get("study_name"),
            direction=optuna_confs.get("direction"),
        )

        if not suggested_conf:
            suggested_conf = {}
        self.suggested_conf = suggested_conf
        self.suggested_optimizers = suggested_conf.get(
            "optimizer", self.DEFAULT_OPTIMIZERS
        )

        default_lr = self.DEFAULT_LR
        self.suggested_lr = (
            suggested_conf.get("lr")
            or suggested_conf.get("learning_rate")
            or default_lr
        )

    @abstractmethod
    def get_trial_model(self, trial) -> Dict[str, str]:
        raise NotImplementedError

    @abstractmethod
    def build_model(self, **kwargs) -> torch.nn.Module:
        raise NotImplementedError

    # noinspection PyTypeChecker
    def _objective(
        self,
        trial: optuna.Trial,
    ):
        """
        Defines _objective function to minimize
        Args:
            trial [optuna.Trial]: optuna.Trial object passed during `optuna.Study.optimize`

        Returns:

        """
        trainer = pl.Trainer(
            logger=True,
            gpus=1 if torch.cuda.is_available() else None,
            max_epochs=self.max_epochs,
            callbacks=PyTorchLightningPruningCallback(
                trial, monitor=self.optimization_metric
            ),
        )
        trial_confs = self.get_trial_model(trial)
        model = self.build_model(**trial_confs)
        hparams = dict(model=model.hparams)
        trainer.logger.log_hyperparams(hparams)
        trainer.fit(model, datamodule=self.datamodule)

        logger.debug(trainer.callback_metrics)
        return trainer.callback_metrics[self.optimization_metric].item()

    def hp_tune(self):
        """
        Search Hyperparameter and builds model with the best params
        Returns:
            sets `self.model` to the best model.

        """
        self.study.optimize(
            self._objective, n_trials=self.n_trials, timeout=self.timeout
        )
        self.model = self.build_model(**self.study.best_params)



import torch
from flash.text.classification import TextClassifier

from gradsflow.autoclassifier.base import AutoClassifier


# noinspection PyTypeChecker
class AutoTextClassifier(AutoClassifier):
    """
    Automatically finds Text Classification Model

    Args:
        datamodule: PL Lightning DataModule with `num_classes` property.
        max_epochs: default=10.
        n_trials: default=100.
        optimization_metric: Optional[str] = None.
        suggested_backbones: Union[List, str, None] = None.
        suggested_conf [Optional[dict] = None]: This sets Trial suggestions for optimizer,
            learning rate, and all the hyperparameters.
        timeout: Hyperparameter search will stop after timeout.

    Example:
        ```python
            suggested_conf = dict(
                optimizers=["adam"],
                lr=(5e-4, 1e-3),
            )
            model = AutoTextClassifier(datamodule,
                                       suggested_backbones=['sgugger/tiny-distilbert-classification'],
                                       suggested_conf=suggested_conf,
                                       max_epochs=1,
                                       optimization_metric="val_accuracy",
                                       timeout=30)
            model.hp_tune()
        ```
    """

    DEFAULT_BACKBONES = [
        "distilbert-base-uncased-finetuned-sst-2-english",
        "sgugger/tiny-distilbert-classification",
    ]

    def build_model(self, **kwargs) -> torch.nn.Module:
        backbone = kwargs["backbone"]
        optimizer = kwargs["optimizer"]
        learning_rate = kwargs["lr"]

        return TextClassifier(
            self.num_classes,
            backbone=backbone,
            optimizer=self.OPTIMIZER_INDEX[optimizer],
            learning_rate=learning_rate,
        )



import pytest
import torch
from flash.image import ImageClassificationData, ImageClassifier

from gradsflow.autoclassifier import AutoImageClassifier

datamodule = ImageClassificationData.from_folders(
    train_folder="data/hymenoptera_data/train/",
    val_folder="data/hymenoptera_data/val/",
)


def test_forward():
    model = AutoImageClassifier(datamodule)

    with pytest.raises(UserWarning):
        model.forward(torch.rand(1, 3, 8, 8))


def test_build_model():
    model = AutoImageClassifier(
        datamodule,
        max_epochs=1,
        timeout=5,
        suggested_backbones="ssl_resnet18",
        n_trials=1,
    )
    kwargs = {"backbone": "ssl_resnet18", "optimizer": "adam", "lr": 1e-1}
    model.model = model.build_model(**kwargs)
    assert isinstance(model.model, ImageClassifier)


def test_model():
    model = AutoImageClassifier(
        datamodule,
        max_epochs=1,
        timeout=5,
        suggested_backbones="ssl_resnet18",
        n_trials=1,
    )
    model.hp_tune()



from flash.core.data.utils import download_data
from flash.image import ImageClassificationData

from gradsflow.autoclassifier import AutoImageClassifier

# 1. Create the DataModule
download_data("https://pl-flash-data.s3.amazonaws.com/hymenoptera_data.zip", "./data")

datamodule = ImageClassificationData.from_folders(
    train_folder="data/hymenoptera_data/train/",
    val_folder="data/hymenoptera_data/val/",
)

model = AutoImageClassifier(
    datamodule, max_epochs=2, optimization_metric="val_accuracy"
)
print("AutoImageClassifier initialised!")

model.fit()



import torch

from gradsflow.utility.common import (
    module_to_cls_index,
    download,
    get_file_extension,
    get_files,
    listify,
)


def test_create_module_index():
    assert isinstance(module_to_cls_index(torch.optim), dict)


def test_get_files():
    assert len(get_files("./")) != 0


def test_get_file_extension():
    assert get_file_extension("image.1.png") == "png"


def test_download():
    assert "gradsflow" in (download("README.md")).lower()


def test_listify():
    assert listify(None) == []
    assert listify(1) == [1]
    assert listify((1, 2)) == [1, 2]



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['gradsflow',
 'gradsflow.autoclassifier',
 'gradsflow.automodel',
 'gradsflow.utility']

package_data = \
{'': ['*']}

install_requires = \
['optuna==2.9.1',
 'smart_open==5.1.0',
 'lightning-flash[all]==0.4.0',
 'pytorch-lightning==1.4.0',
 'loguru==0.5']

extras_require = \
{'test': ['pytest', 'coverage']}

setup(name='gradsflow',
      version='0.0.1b1',
      description='An AutoML Library made with Optuna and PyTorch Lightning',
      author='Aniket Maurya',
      author_email='hello@gradsflow.com',
      url='https://github.com/gradsflow/gradsflow',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



from unittest.mock import MagicMock

from gradsflow.autoclassifier import AutoTextClassifier


def test_build_model():
    suggested_conf = dict(
        optimizers=["adam"],
        lr=(5e-4, 1e-3),
    )
    datamodule = MagicMock()
    datamodule.num_classes = 2
    model = AutoTextClassifier(
        datamodule,
        suggested_backbones=["sgugger/tiny-distilbert-classification"],
        suggested_conf=suggested_conf,
        max_epochs=1,
        optimization_metric="val_accuracy",
        timeout=5,
        n_trials=1,
    )

    model_confs = {
        "backbone": model.DEFAULT_BACKBONES[-1],
        "optimizer": "adam",
        "lr": 1e-3,
    }
    model.build_model(**model_confs)
